@@ -28,7 +28,13 @@
 #include "layoutreader.h"
 #include <gtk/gtk.h>
 #include <gdk/gdkx.h>
+#ifdef ENABLE_AT_SPI2
+#define AT_SPI
+#include <dbus/dbus.h>
+#include <atspi/atspi.h>
+#endif
 #ifdef ENABLE_AT_SPI
+#define AT_SPI
 #include <cspi/spi.h>
 #endif
 
@@ -44,14 +50,18 @@ void flo_destroy (void)
 	//gtk_exit (0);
 }
 
-#ifdef ENABLE_AT_SPI
+#ifdef AT_SPI
 /* Called to destroy the icon */
 void flo_icon_destroy (GtkWidget *widget, gpointer user_data)
 {
 	struct florence *florence=(struct florence *)user_data;
 	if (florence->icon) gtk_object_destroy(GTK_OBJECT(florence->icon));
 	if (florence->obj) {
+#ifdef ENABLE_AT_SPI2
+		g_object_unref(florence->obj);
+#else
 		Accessible_unref(florence->obj);
+#endif
 		florence->obj=NULL;
 	}
 	florence->icon=NULL;
@@ -65,7 +75,11 @@ void flo_icon_press (GtkWidget *window, GdkEventButton *event, gpointer user_dat
 	florence->icon=NULL;
 	view_show(florence->view, florence->obj);
 	if (florence->obj) {
+#ifdef ENABLE_AT_SPI2
+		g_object_unref(florence->obj);
+#else
 		Accessible_unref(florence->obj);
+#endif
 		florence->obj=NULL;
 	}
 }
@@ -73,9 +87,11 @@ void flo_icon_press (GtkWidget *window, GdkEventButton *event, gpointer user_dat
 /* on expose event: display florence icon */
 void flo_icon_expose (GtkWidget *window, GdkEventExpose* pExpose, void *userdata)
 {
-	cairo_t *context;
+	cairo_t *context, *mask_context;
 	RsvgHandle *handle;
 	GError *error=NULL;
+	gdouble w, h;
+	GdkBitmap *mask=NULL;
 
 	context=gdk_cairo_create(window->window);
 	cairo_set_operator(context, CAIRO_OPERATOR_SOURCE);
@@ -83,12 +99,24 @@ void flo_icon_expose (GtkWidget *window, GdkEventExpose* pExpose, void *userdata
 	handle=rsvg_handle_new_from_file(ICONDIR "/florence.svg", &error);
 	if (error) flo_error(_("Error loading florence icon: %s"), error->message);
 	else {
+		w=settings_double_get("window/scalex")*2;
+		h=settings_double_get("window/scaley")*2;
+		if (!(mask=(GdkBitmap*)gdk_pixmap_new(NULL, w, h, 1)))
+			flo_fatal(_("Unable to create mask"));
+		mask_context=gdk_cairo_create(mask);
+		cairo_set_source_rgba(mask_context, 0.0, 0.0, 0.0, 0.0);
+		cairo_set_operator(mask_context, CAIRO_OPERATOR_SOURCE);
+		cairo_paint(mask_context);
+		cairo_set_operator(mask_context, CAIRO_OPERATOR_OVER);
+		style_render_svg(mask_context, handle, w, h, TRUE, NULL);
+		gdk_window_shape_combine_mask(window->window, mask, 0, 0);
 		cairo_set_source_rgba(context, 0.0, 0.0, 0.0, 100.0);
 		cairo_set_operator(context, CAIRO_OPERATOR_DEST_OUT);
 		cairo_paint(context);
 		cairo_set_operator(context, CAIRO_OPERATOR_OVER);
-		style_render_svg(context, handle, settings_double_get("window/zoom")*2,
-			settings_double_get("window/zoom")*2, TRUE, NULL);
+		style_render_svg(context, handle, w, h, TRUE, NULL);
+		cairo_destroy(mask_context);
+		g_object_unref(G_OBJECT(mask));
 		rsvg_handle_free(handle);
 	}
 
@@ -97,18 +125,26 @@ void flo_icon_expose (GtkWidget *window, GdkEventExpose* pExpose, void *userdata
 
 /* Show an intermediate icon before showing the keyboard (if intermediate_icon is activated) 
  * otherwise, directly show the keyboard */
+#ifdef ENABLE_AT_SPI2
+void flo_check_show (struct florence *florence, AtspiAccessible *obj)
+#else
 void flo_check_show (struct florence *florence, Accessible *obj)
+#endif
 {
 	if (GTK_WIDGET_VISIBLE(florence->view->window)) view_hide(florence->view);
 	if (settings_get_bool("behaviour/intermediate_icon")) {
+#ifdef ENABLE_AT_SPI2
+		if (florence->obj) g_object_unref(florence->obj);
+#else
 		if (florence->obj) Accessible_unref(florence->obj);
+#endif
 		florence->obj=obj;
 		if (!florence->icon) {
 			florence->icon=GTK_WINDOW(gtk_window_new(GTK_WINDOW_TOPLEVEL));
 			gtk_window_set_keep_above(florence->icon, TRUE);
 			gtk_window_set_skip_taskbar_hint(florence->icon, TRUE);
-			gtk_widget_set_size_request(GTK_WIDGET(florence->icon), settings_double_get("window/zoom")*2,
-				settings_double_get("window/zoom")*2);
+			gtk_widget_set_size_request(GTK_WIDGET(florence->icon), settings_double_get("window/scalex")*2,
+				settings_double_get("window/scaley")*2);
 			gtk_container_set_border_width(GTK_CONTAINER(florence->icon), 0);
 			gtk_window_set_decorated(florence->icon, FALSE);
 			gtk_window_set_position(florence->icon, GTK_WIN_POS_MOUSE);
@@ -130,15 +166,30 @@ void flo_check_show (struct florence *florence, Accessible *obj)
 
 /* Called when a widget is focused.
  * Check if the widget is editable and show the keyboard or hide if not. */
+#ifdef ENABLE_AT_SPI2
+void flo_focus_event (const AtspiEvent *event, void *user_data)
+#else
 void flo_focus_event (const AccessibleEvent *event, void *user_data)
+#endif
 {
 	struct florence *florence=(struct florence *)user_data;
 	gboolean hide=FALSE;
 
+#ifdef ENABLE_AT_SPI2
+	if (atspi_accessible_get_role(event->source, NULL)==ATSPI_ROLE_TERMINAL ||
+		(((atspi_accessible_get_role(event->source, NULL)==ATSPI_ROLE_TEXT) ||
+		(atspi_accessible_get_role(event->source, NULL)==ATSPI_ROLE_PASSWORD_TEXT)) &&
+		 atspi_state_set_contains(atspi_accessible_get_state_set(event->source), ATSPI_STATE_EDITABLE))) {
+#else
 	if (Accessible_getRole(event->source)==SPI_ROLE_TERMINAL || Accessible_isEditableText(event->source)) {
+#endif
 		if (event->detail1) {
 			flo_check_show(florence, event->source);
+#ifdef ENABLE_AT_SPI2
+			g_object_ref(event->source);
+#else
 			Accessible_ref(event->source);
+#endif
 		} else {
 			hide=TRUE;
 		}
@@ -148,39 +199,72 @@ void flo_focus_event (const AccessibleEvent *event, void *user_data)
 	if (hide) {
 		view_hide(florence->view);
 		if (florence->icon) gtk_object_destroy(GTK_OBJECT(florence->icon));
+#ifdef ENABLE_AT_SPI2
+		if (florence->obj) g_object_unref(florence->obj);
+#else
 		if (florence->obj) Accessible_unref(florence->obj);
+#endif
 		florence->icon=NULL;
 		florence->obj=NULL;
 	}
 }
+#endif
 
+#ifdef AT_SPI
 /* Shouldn't be used but it seems like we need to traverse accessible widgets when a new window is open to trigger
  * focus events. This is at least the case for gedit. Will need to check how this all work.
  * Can't we get a focus event when the widget is greated focussed? */
+#ifdef ENABLE_AT_SPI2
+void flo_traverse (struct florence *florence, AtspiAccessible *obj)
+#else
 void flo_traverse (struct florence *florence, Accessible *obj)
+#endif
 {
 	int n_children, i;
-	Accessible *child;
+#ifdef ENABLE_AT_SPI2
+	AtspiAccessible *child;
+	n_children=atspi_accessible_get_child_count(obj, NULL);
 
+	if (!atspi_accessible_get_table(obj)) {
+#else
+	Accessible *child;
 	n_children=Accessible_getChildCount(obj);
-	if (!Accessible_isTable(obj))
-	{
+
+	if (!Accessible_isTable(obj)) {
+#endif
 		for (i=0;i<n_children;++i)
 		{
+#ifdef ENABLE_AT_SPI2
+			child=atspi_accessible_get_child_at_index(obj, i, NULL);
+			if (atspi_state_set_contains(atspi_accessible_get_state_set(child), ATSPI_STATE_FOCUSED) &&
+				(atspi_accessible_get_role(child, NULL)==ATSPI_ROLE_TERMINAL ||
+				(((atspi_accessible_get_role(child, NULL)==ATSPI_ROLE_TEXT) ||
+				(atspi_accessible_get_role(child, NULL)==ATSPI_ROLE_PASSWORD_TEXT)) &&
+				 atspi_state_set_contains(atspi_accessible_get_state_set(child), ATSPI_STATE_EDITABLE)))) {
+#else
 			child=Accessible_getChildAtIndex(obj, i);
 			if (Accessible_isEditableText(child) &&
 				AccessibleStateSet_contains(Accessible_getStateSet(child), SPI_STATE_FOCUSED)) {
+#endif
 				flo_check_show(florence, child);
 			} else {
 				flo_traverse(florence, child);
+#ifdef ENABLE_AT_SPI2
+				if (child) g_object_unref(child);
+#else
 				if (child) Accessible_unref(child);
+#endif
 			}
 		}
 	}
 }
 
 /* Called when a window is created */
+#ifdef ENABLE_AT_SPI2
+void flo_window_create_event (const AtspiEvent *event, gpointer user_data)
+#else
 void flo_window_create_event (const AccessibleEvent *event, gpointer user_data)
+#endif
 {
 	/* For some reason, focus state change does happen after traverse 
 	 * ==> did I misunderstand? */
@@ -197,20 +281,26 @@ void flo_switch_mode (struct florence *florence, gboolean auto_hide)
 #ifdef ENABLE_AT_SPI
 	static AccessibleEventListener *focus_listener=NULL;
 	static AccessibleEventListener *window_listener=NULL;
-	int i;
 	Accessible *obj=NULL;
 #endif
+#ifdef AT_SPI
+	int i;
+#endif
 
 	if (auto_hide) {
 		if (!status_spi_is_enabled(florence->status)) {
 			flo_warn(_("SPI is disabled: Unable to switch auto-hide mode on."));
 			return;
 		}
-#ifdef ENABLE_AT_SPI
+#ifdef AT_SPI
 		view_hide(florence->view);
-		focus_listener=SPI_createAccessibleEventListener (flo_focus_event, (void*)florence);
+#ifdef ENABLE_AT_SPI2
+		atspi_event_listener_register_from_callback(flo_focus_event, (void*)florence, NULL, "object:state-changed:focused", NULL);
+		atspi_event_listener_register_from_callback(flo_window_create_event, (void*)florence, NULL, "window:activate", NULL);
+#else
+		focus_listener=SPI_createAccessibleEventListener(flo_focus_event, (void*)florence);
 		SPI_registerGlobalEventListener(focus_listener, "object:state-changed:focused");
-		window_listener=SPI_createAccessibleEventListener (flo_window_create_event, (void*)florence);
+		window_listener=SPI_createAccessibleEventListener(flo_window_create_event, (void*)florence);
 		SPI_registerGlobalEventListener(window_listener, "window:activate");
 		for (i=1;i<=SPI_getDesktopCount();i++) {
 			obj=SPI_getDesktop(i);
@@ -219,7 +309,17 @@ void flo_switch_mode (struct florence *florence, gboolean auto_hide)
 				Accessible_unref(obj);
 			}
 		}
+#endif
 	} else {
+#ifdef ENABLE_AT_SPI2
+		if (!atspi_event_listener_deregister_from_callback(flo_focus_event, (void*)florence, "object:state-changed:focused", NULL)) {
+			flo_warn(_("AT SPI: problem deregistering focus listener"));
+		}
+		if (!atspi_event_listener_deregister_from_callback(flo_window_create_event, (void*)florence, "window:activate", NULL)) {
+			flo_warn(_("AT SPI: problem deregistering window listener"));
+		}
+#endif
+#ifdef ENABLE_AT_SPI
 		if (focus_listener) {
 			SPI_deregisterGlobalEventListenerAll(focus_listener);
 			AccessibleEventListener_unref(focus_listener);
@@ -230,9 +330,12 @@ void flo_switch_mode (struct florence *florence, gboolean auto_hide)
 			AccessibleEventListener_unref(window_listener);
 			window_listener=NULL;
 		}
+#endif
 		view_show(florence->view, NULL);
 		if (florence->icon) gtk_object_destroy(GTK_OBJECT(florence->icon));
+#ifdef ENABLE_AT_SPI
 		if (florence->obj) Accessible_unref(obj);
+#endif
 		florence->obj=NULL;
 		florence->icon=NULL;
 #endif
@@ -294,6 +397,9 @@ gboolean flo_mouse_leave_event (GtkWidget *window, GdkEvent *event, gpointer use
 /* handles button press events */
 gboolean flo_button_press_event (GtkWidget *window, GdkEventButton *event, gpointer user_data)
 {
+#ifdef ENABLE_RAMBLE
+	gchar *val;
+#endif
 	struct florence *florence=(struct florence *)user_data;
 	struct key *key=NULL;
 	
@@ -311,7 +417,8 @@ gboolean flo_button_press_event (GtkWidget *window, GdkEventButton *event, gpoin
 	}
 
 #ifdef ENABLE_RAMBLE
-	if (!strcmp("ramble", settings_get_string("behaviour/input_method"))) {
+	val=settings_get_string("behaviour/input_method");
+	if (!strcmp("ramble", val)) {
 		ramble_start(florence->ramble);
 	} else {
 #endif
@@ -319,6 +426,7 @@ gboolean flo_button_press_event (GtkWidget *window, GdkEventButton *event, gpoin
 	status_timer_stop(florence->status);
 #ifdef ENABLE_RAMBLE
 	}
+	if (val) g_free(val);
 #endif
 	return FALSE;
 }
@@ -359,9 +467,26 @@ gboolean flo_timer_update(gpointer data)
 /* bring the window back to front: to be calles periodically */
 gboolean flo_to_top(gpointer data)
 {
-	if (GTK_WIDGET_VISIBLE(GTK_WIDGET(data))) gtk_window_present(GTK_WINDOW(data));
+	struct florence *florence=data;
+	GtkWindow *window=GTK_WINDOW(view_window_get(florence->view));
+	if (!settings_get_bool("window/keep_on_top")) return FALSE;
+	if (GTK_WIDGET_VISIBLE(GTK_WIDGET(window))) gtk_window_present(window);
 	return TRUE;
 }
+/* start keeping the keyboard back to front every second */
+void flo_start_keep_on_top(struct florence *florence, gboolean keep_on_top)
+{
+	if (settings_get_bool("window/keep_on_top")) {
+		g_timeout_add(FLO_TO_TOP_TIMEOUT, flo_to_top, florence);
+	}
+}
+
+/* Triggered by gconf when the "keep_on_top" parameter is changed. */
+void flo_set_keep_on_top(GConfClient *client, guint xnxn_id, GConfEntry *entry, gpointer user_data)
+{
+	struct florence *florence=user_data;
+	flo_start_keep_on_top(florence, gconf_value_get_bool(gconf_entry_get_value(entry)));
+}
 
 /* handles mouse motion events 
  * update the keyboard key under the mouse */
@@ -369,7 +494,9 @@ gboolean flo_mouse_move_event(GtkWidget *window, GdkEvent *event, gpointer user_
 {
 #ifdef ENABLE_RAMBLE
 	enum key_hit hit;
+	gchar *algo;
 #endif
+	gchar *val;
 	struct florence *florence=(struct florence *)user_data;
 	if (status_get_moving(florence->status)) {
 		gtk_window_move(GTK_WINDOW(window), (gint)((GdkEventMotion*)event)->x_root-florence->xpos,
@@ -378,16 +505,19 @@ gboolean flo_mouse_move_event(GtkWidget *window, GdkEvent *event, gpointer user_
 		/* Remember mouse position for moving */
 		florence->xpos=(gint)((GdkEventMotion*)event)->x;
 		florence->ypos=(gint)((GdkEventMotion*)event)->y;
+		val=settings_get_string("behaviour/input_method");
 #ifdef ENABLE_RAMBLE
 		struct key *key=status_hit_get(florence->status, florence->xpos, florence->ypos, &hit);
-		if (!strcmp("ramble", settings_get_string("behaviour/input_method"))) {
+		if (!strcmp("ramble", val)) {
 			florence->view->ramble=florence->ramble;
+			algo=settings_get_string("behaviour/ramble_algo");
 			if ((hit==KEY_BORDER) &&
 				(status_focus_get(florence->status)==key) &&
-				(!strcmp("time", settings_get_string("behaviour/ramble_algo")))) {
+				(!strcmp("time", algo))) {
 				ramble_time_reset(florence->ramble);
 				status_focus_set(florence->status, NULL);
 			}
+			if (algo) g_free(algo);
 			if (ramble_started(florence->ramble) &&
 				ramble_add(florence->ramble, GTK_WIDGET(florence->view->window)->window,
 					florence->xpos, florence->ypos, key)) {
@@ -403,11 +533,12 @@ gboolean flo_mouse_move_event(GtkWidget *window, GdkEvent *event, gpointer user_
 #endif
 		if (status_focus_get(florence->status)!=key) {
 			if (key && settings_double_get("behaviour/timer")>0.0 &&
-				(!strcmp("timer", settings_get_string("behaviour/input_method")))) {
+				(!strcmp("timer", val))) {
 				status_timer_start(florence->status, flo_timer_update, (gpointer)florence);
 			} else status_timer_stop(florence->status);
 			status_focus_set(florence->status, key);
 		}
+		if (val) g_free(val);
 	}
 	return FALSE;
 }
@@ -502,7 +633,7 @@ struct florence *flo_new(gboolean gnome, const gchar *focus_back)
 struct florence *flo_new(gboolean gnome, const gchar *focus_back, PanelApplet *applet)
 #endif
 {
-  struct florence *florence=(struct florence *)g_malloc(sizeof(struct florence));
+	struct florence *florence=(struct florence *)g_malloc(sizeof(struct florence));
 	if (!florence) flo_fatal(_("Unable to allocate memory for florence"));
 	memset(florence, 0, sizeof(struct florence));
 
@@ -515,11 +646,13 @@ struct florence *flo_new(gboolean gnome, const gchar *focus_back, PanelApplet *a
 #endif
 
 	florence->status=status_new(focus_back);
-#ifdef ENABLE_AT_SPI
-	if (0!=SPI_init()) {
-		if ((!gnome) || (!flo_check_at_spi())) {
+#ifdef AT_SPI
+#ifdef ENABLE_AT_SPI2
+	if (atspi_init()) {
+#else
+	if (SPI_init() && ((!gnome) || (!flo_check_at_spi()))) {
+#endif
 			status_spi_disable(florence->status);
-		}
 	}
 #else
 	flo_warn(_("AT-SPI has been disabled at compile time: auto-hide mode is disabled."));
@@ -533,7 +666,7 @@ struct florence *flo_new(gboolean gnome, const gchar *focus_back, PanelApplet *a
 	florence->view=view_new(florence->status, florence->style, florence->keyboards);
 #endif
 	status_view_set(florence->status, florence->view);
-	if (settings_get_bool("window/keep_on_top")) g_timeout_add(FLO_TO_TOP_TIMEOUT, flo_to_top, view_window_get(florence->view));
+	flo_start_keep_on_top(florence, settings_get_bool("window/keep_on_top"));
 
 	g_signal_connect(G_OBJECT(view_window_get(florence->view)), "destroy", G_CALLBACK(flo_destroy), NULL);
 	g_signal_connect(G_OBJECT(view_window_get(florence->view)), "motion-notify-event",
@@ -545,10 +678,12 @@ struct florence *flo_new(gboolean gnome, const gchar *focus_back, PanelApplet *a
 	g_signal_connect(G_OBJECT(view_window_get(florence->view)), "button-release-event",
 		G_CALLBACK(flo_button_release_event), florence);
 #ifndef APPLET
-	flo_switch_mode(florence, settings_get_bool("behaviour/auto_hide"));
+	if (settings_get_bool("behaviour/hide_on_start") && (!settings_get_bool("behaviour/auto_hide"))) view_hide(florence->view);
+	else flo_switch_mode(florence, settings_get_bool("behaviour/auto_hide"));
 	florence->trayicon=trayicon_new(florence->view, G_CALLBACK(flo_destroy));
 #endif
 	settings_changecb_register("behaviour/auto_hide", flo_set_auto_hide, florence);
+	settings_changecb_register("window/keep_on_top", flo_set_keep_on_top, florence);
 	/* TODO: just reload the style, no need to reload the whole layout */
 	settings_changecb_register("layout/style", flo_layout_reload, florence);
 	settings_changecb_register("layout/file", flo_layout_reload, florence);
@@ -559,9 +694,13 @@ struct florence *flo_new(gboolean gnome, const gchar *focus_back, PanelApplet *a
 /* liberate all the memory used by florence */
 void flo_free(struct florence *florence)
 {
-#ifdef ENABLE_AT_SPI
+#ifdef AT_SPI
+#ifdef ENABLE_AT_SPI2
+	atspi_exit();
+#else
 	SPI_exit();
 #endif
+#endif
 #ifndef APPLET
 	trayicon_free(florence->trayicon);
 #endif
